// audio functions for butt
//
// Copyright 2007 by Daniel Noethen.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//


#include <stdio.h>
#include <stdlib.h>

#include <string.h>

#include "butt.h"
#include "cfg.h"
#include "port_audio.h"
#include "parseconfig.h"
#include "lame_encode.h"
#include "shoutcast.h"

#include "config.h"


#define PA_FRAMES 1024
#define FRAMES_MUL 8
#define BUF_FRAMES PA_FRAMES*FRAMES_MUL

char* encode_buf;
short *pa_pcm_buf;
short *pa_vu_buf;
int buf_index;
int buf_size;
int framepacket_size;

int try_to_connect;
bool pa_new_frames;
bool snd_xrun;
bool pa_new_buf;
bool reconnect;
int pa_callback_counter;

PaStream *stream;

int snd_init()
{

    char info[100];

    PaError p_err;
    if(( p_err = Pa_Initialize()) != paNoError )
    {
        snprintf(info, 100, "PortAudio init failed:\n%s\n",
                            Pa_GetErrorText(p_err));
        ALERT(info);

        return 1;
    }
    reconnect = 0;
    buf_index = 0;

    return 0;
}

int snd_open_stream()
{

    int samplerate;

    PaDeviceIndex pa_dev_id;
    PaStreamParameters pa_params;
    PaError pa_err;
    const PaDeviceInfo *pa_dev_info;
    char info_text[128];

    
    
    
    pa_callback_counter = 0;

    buf_size = BUF_FRAMES * 2;
    framepacket_size = PA_FRAMES * cfg.audio.channel;

    pa_vu_buf  = (short*)malloc(framepacket_size * sizeof(short));
    pa_pcm_buf = (short*)malloc(buf_size * sizeof(short));
    encode_buf = (char*)malloc(buf_size * sizeof(char));

    samplerate = cfg.audio.samplerate;

    

    if(cfg.audio.pcm_list[cfg.audio.dev_num]->dev_id == -1)
        pa_dev_id = Pa_GetDefaultInputDevice();
    else
        pa_dev_id = cfg.audio.pcm_list[cfg.audio.dev_num]->dev_id;

    pa_dev_info = Pa_GetDeviceInfo(pa_dev_id);

    pa_params.device = pa_dev_id;
    pa_params.channelCount = cfg.audio.channel;
    pa_params.sampleFormat = paInt16;
    pa_params.suggestedLatency = pa_dev_info->defaultHighInputLatency;
    pa_params.hostApiSpecificStreamInfo = NULL;

    sprintf(info_text, "Trying to open sound device:\n%s",
                       pa_dev_info->name);

    print_info(info_text, 0);

    pa_err = Pa_IsFormatSupported(&pa_params, NULL, samplerate);
    if(pa_err != paFormatIsSupported)
    {
        if(pa_err == paInvalidSampleRate)
        {
            sprintf(info_text,
                    "Samplerate not supported: %d\n"
                    "Trying default samplerate: %d",
                    samplerate, (int)pa_dev_info->defaultSampleRate);

            print_info(info_text, 1);
            if(Pa_IsFormatSupported(&pa_params, NULL,
                        pa_dev_info->defaultSampleRate) != paFormatIsSupported)
            {
                print_info("FAILED", 1);
                return 1;

            }
            else
            {
                samplerate = (int)pa_dev_info->defaultSampleRate;
                print_info("SUCCEED", 0);
            }
        }
        else
        {
            sprintf(info_text, "Format not supported. \n%s\n", Pa_GetErrorText(pa_err));
            print_info(info_text, 1);
            return 1;
        }
    }
    else
    {

        print_info("SUCCEED", 0);
    }


    pa_err = Pa_OpenStream(&stream, &pa_params, NULL,
                        samplerate, PA_FRAMES,
                        paNoFlag, snd_callback, NULL);

    if(pa_err != paNoError)
    {
        printf("error opening sound device: \n%s\n", Pa_GetErrorText(pa_err));
        return 1;

    }

    Pa_StartStream(stream);

    return 0;
}


int snd_send_buf()
{
    int encode_bytes_read = 0;
    int (*xc_send)(char *buf, int buf_len) = NULL;
    short *local_audio_buf = (short*)malloc(buf_size*sizeof(short));

    pa_callback_counter = 0;

    if(cfg.srv[cfg.selected_srv]->type == SHOUTCAST)
    {
            xc_send = &sc_send;
    }
    if(cfg.srv[cfg.selected_srv]->type == ICECAST)
    {

#if HAVE_LIBSHOUT
        xc_send = &ic_send;
#else
        pa_new_buf = 0;
        free(local_audio_buf);
        return 1;
#endif
    }
    
    //Save the collected audio data into a local buf before sending it
    //This prevents that the callbackfunction could override some bytes
    memcpy(local_audio_buf, pa_pcm_buf, buf_size*sizeof(short));

    if(record)
    {
        encode_bytes_read = lame_enc_buf(local_audio_buf, encode_buf, 
                                BUF_FRAMES, gfp_record, buf_size);
        fwrite(encode_buf, 1, encode_bytes_read, cfg.rec.fd);

    }
    if(!strcmp(cfg.audio.codec, "mp3"))
        encode_bytes_read = lame_enc_buf(local_audio_buf, encode_buf, 
                                BUF_FRAMES, gfp_stream, buf_size);
#if HAVE_LIBVORBIS
    else
        encode_bytes_read = vorbis_enc_buf(local_audio_buf,
                                                  encode_buf,
                                                  BUF_FRAMES);
#endif
    if(xc_send(encode_buf, encode_bytes_read) == -1)
    {
        free(local_audio_buf);
        pa_new_buf = 0;
        return 1; //disconnected
    }
    
    network_bytes_sent += encode_bytes_read;
    pa_new_buf = 0;
    free(local_audio_buf);

    return 0;

}

void snd_update_vu() //Shows every 1024 recorded frames on the vu-meter
{
    for(int i = 0; i < framepacket_size; i += cfg.audio.channel)
    {
        vu_meter(pa_vu_buf[i], pa_vu_buf[i+(cfg.audio.channel-1)]);
    }
    pa_new_frames = 0;
}


int snd_callback(const void *input, //this function is called by PortAudio
                 void *output,
                 unsigned long frameCount,
                 const PaStreamCallbackTimeInfo* timeInfo,
                 PaStreamCallbackFlags statusFlags,
                 void *userData)
{
    //If this function gets more often called than FRAMES_MUL,
    //then the snd_send_buf function wasn fast enough to send the data to the server
    //this results in a buffer overrun
    if(pa_callback_counter++ >= FRAMES_MUL)
       snd_xrun = 1;
    else
       snd_xrun = 0;

    memcpy(pa_pcm_buf+buf_index*framepacket_size, input, framepacket_size*sizeof(short));
    memcpy(pa_vu_buf, input, framepacket_size*sizeof(short));

    buf_index++;
    if(buf_index == FRAMES_MUL)
    {
        pa_new_buf = 1;
        buf_index = 0;
    }

    pa_new_frames = 1;

    return 0;
}

snd_dev_t **snd_get_devices(int *dev_count)
{

    int devcount, i, dev_num = 1;
    const PaDeviceInfo *p_di;
    char *p;

    snd_dev_t **dev_list;

    dev_list = (snd_dev_t**)malloc(100*sizeof(snd_dev_t*));

    for(i = 0; i < 100; i++)
        dev_list[i] = (snd_dev_t*)malloc(sizeof(snd_dev_t));


    dev_list[0]->name = (char*) malloc(strlen("Default PCM device (default)")+1);
    sprintf(dev_list[0]->name, "Default PCM device (default)");
    dev_list[0]->dev_id = -1;

    devcount = Pa_GetDeviceCount();
    for( i = 0; i < devcount && i < 100; i++ )
    {
        p_di = Pa_GetDeviceInfo(i);

        //Save only devices with input Channels
        if(p_di->maxInputChannels > 0)
        {
            dev_list[dev_num]->name = (char*) malloc(strlen(p_di->name)+1);
            sprintf(dev_list[dev_num]->name, p_di->name);
            dev_list[dev_num]->dev_id = i;

            while((p = strchr(dev_list[dev_num]->name, '/')) != NULL)
            {
                //replace every '/' with an '-'
                //otherweise FLTK will create a submenu for every '/' in the choice menu :-(
                *p = '-';
            }
            dev_num++;
        }
    }

    *dev_count = dev_num;


    return dev_list;

}


void snd_close()
{

    Pa_StopStream(stream);
    Pa_CloseStream(stream);
    Pa_Terminate();

    free(pa_vu_buf);
    free(pa_pcm_buf);
    free(encode_buf);
}


